/* 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/*
 * Tcl support by Alex Badea (vampire@cs.callatis.ro)
 * (EXPERIMENTAL)
 */

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>

#include <tcl.h>
#include <glib.h>

#include "xchat.h"
#include "plugin.h"

#define MSG_PREFIX "\0039tcl:\003 "

char *mod_name = "TCL";
char *mod_desc = "Loadable TCL scripting engine";

extern struct module *module_find(char *name);
extern void unhook_signal(struct xp_signal *sig);
extern void module_add_cmds(struct module_cmd_set *cmds);
extern int hook_signal(struct xp_signal *sig);

extern struct session *current_tab;
extern int is_session(struct session *sess);
extern void PrintText(struct session *sess, unsigned char *text);

extern int is_server(struct server *serv);
extern int tcp_send(struct server *serv, char *buf);


static Tcl_Interp *tcl_new_interp(void);
static void tcl_kill_interp(Tcl_Interp *interp);

#ifdef __GNUC__
static void print_text(struct session *sess, char *fmt, ...)
__attribute__ ((format (printf, 2, 3)));
static void Tcl_SetResultFmt(Tcl_Interp *interp, char *fmt, ...)
__attribute__ ((format (printf, 2, 3)));
#else
static void print_text(struct session *sess, char *fmt, ...);
static void Tcl_SetResultFmt(Tcl_Interp *interp, char *fmt, ...);
#endif

static void print_text(struct session *sess, char *fmt, ...)
{
	va_list args;
	char buf[4096];

	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	va_end(args);
	PrintText(sess, buf);
}

static void Tcl_SetResultFmt(Tcl_Interp *interp, char *fmt, ...)
{
	va_list args;
	char buf[4096];

	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	va_end(args);
	Tcl_SetResult(interp, buf, TCL_VOLATILE);
}

/*****************************************************************************/
/* Tcl streams */

static int mystream_close(ClientData instanceData, Tcl_Interp *interp)
{
	if (interp)
		interp->result = 0;
	return 0;
}

static int mystream_input(ClientData instanceData, char *buf, int bufSize, int *errorCodePtr)
{
	if (errorCodePtr)
		*errorCodePtr = EINVAL;
	return -1;
}

static int mystream_output(ClientData instanceData, char *buf, int toWrite, int *errorCodePtr)
{
	write(1, buf, toWrite);
	return 0;
}

static void mystream_watch(ClientData instanceData, int mask) {}

static int mystream_get_handle(ClientData instanceData, int direction, ClientData *handlePtr)
{
	return TCL_ERROR;
}

static Tcl_ChannelType mystream_type = {
	"mystream",		/* type name */
	NULL,			/* blockModeProc */
	mystream_close,		/* closeProc */
	mystream_input,		/* inputProc */
	mystream_output,	/* ouputProc */
	NULL,			/* seekProc */
	NULL,			/* setOptionProc */
	NULL,			/* getOptionProc */
	mystream_watch,		/* watchProc */
	mystream_get_handle	/* getHandleProc */
};

/*****************************************************************************/
/* Tcl hook */

#define _SESS	1 /* session */
#define _SERV	2 /* server */
#define _STR	3 /* string */
#define _STAB	4 /* string vector (not supported yet) */

/* *INDENT-OFF* */

static struct signal_map_t {
	char *name;
	int number;
	int num_args;
	int arg_types[5];
} signal_map[] = {
	{ "XP_USERCOMMAND",	XP_USERCOMMAND,	4, { _SESS, _STR, _STAB, _STAB } },
	{ "XP_PRIVMSG",		XP_PRIVMSG,	4, { _SERV, _STR, _STR, _STR } },
	{ "XP_CHANACTION",	XP_CHANACTION,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANMSG",		XP_CHANMSG,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANGENICK",	XP_CHANGENICK,	3, { _SERV, _STR, _STR } },
	{ "XP_JOIN",		XP_JOIN,	4, { _SERV, _STR, _STR } },
	{ "XP_CHANSETKEY",	XP_CHANSETKEY,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANSETLIMIT",	XP_CHANSETLIMIT,4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANOP",		XP_CHANOP,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANVOICE",	XP_CHANVOICE,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANBAN",		XP_CHANBAN,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANRMKEY",	XP_CHANRMKEY,	3, { _SESS, _STR, _STR } },
	{ "XP_CHANRMLIMIT",	XP_CHANRMLIMIT,	3, { _SESS, _STR, _STR } },
	{ "XP_CHANDEOP",	XP_CHANDEOP,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANDEVOICE",	XP_CHANDEVOICE,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANUNBAN",	XP_CHANUNBAN,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANEXEMPT",	XP_CHANEXEMPT,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANRMEXEMPT",	XP_CHANRMEXEMPT,4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANINVITE",	XP_CHANINVITE,	4, { _SESS, _STR, _STR, _STR } },
	{ "XP_CHANRMINVITE",	XP_CHANRMINVITE,4, { _SESS, _STR, _STR, _STR } },
	{ "XP_INBOUND",		XP_INBOUND,	3, { _SESS, _SERV, _STR } },
	{ "XP_HIGHLIGHT",	XP_HIGHLIGHT,	4, { _SESS, _STR, _STR, _STR } },
	{ NULL,			0,		0, {} },
};

/* *INDENT-ON* */

GSList *tcl_plugins = NULL;

struct tcl_plugin {
	char name[64];
	Tcl_Interp *interp;
	GSList *hooks;
};

struct tcl_hook {
	struct tcl_plugin *plugin;
	char cmd_name[64];
	struct xp_signal *sig;
	struct signal_map_t *map_entry;
};

static int tcl_stock_signal_handler(void *a, void *b, void *c, void *d, void *e, char f)
{
	struct tcl_hook *hook;
	char fmt[256], cmd[8192];
	int k;
	char **ptr;

	hook = signal_data;
	if (hook->sig->signal == XP_USERCOMMAND) {
		sprintf(cmd, "%s 0x%08lx %s { ", hook->cmd_name,
			(unsigned long) a, (char *) b);
		strcat(cmd, "\"");
		for (ptr = ((char **) c) + 1; *ptr && **ptr; ) {
			strcat(cmd, *ptr);
			strcat(cmd, "\" ");
			ptr++;
			if (*ptr && **ptr)
				strcat(cmd, "\"");
		}
		strcat(cmd, "} { \"");
		for (ptr = ((char **) d) + 1; *ptr && **ptr; ) {
			strcat(cmd, *ptr);
			strcat(cmd, "\" ");
			ptr++;
			if (*ptr && **ptr)
				strcat(cmd, "\"");
		}
		strcat(cmd, "}");
	} else {
		strcpy(fmt, "%s ");
		for (k = 0; k < hook->map_entry->num_args; k++) {
			switch (hook->map_entry->arg_types[k]) {
			case _SESS:
			case _SERV:
				strcat(fmt, "0x%08lx ");
				break;
			case _STR:
				strcat(fmt, "\"%s\" ");
				break;
			}
		}
		strcat(fmt, "\n");
		snprintf(cmd, sizeof(cmd), fmt, hook->cmd_name, a, b, c, d, e, f);
	}
	if (Tcl_Eval(hook->plugin->interp, cmd) == TCL_OK) {
		int res;
		if (Tcl_GetIntFromObj(hook->plugin->interp, Tcl_GetObjResult(hook->plugin->interp), &res) == TCL_OK)
			return res;
		else
			return 0;
	} else {
		print_text(current_tab, MSG_PREFIX "signal %d in %s: %s",
			hook->sig->signal,
			hook->plugin->name,
			Tcl_GetStringResult(hook->plugin->interp));
		return 0;
	}
}

static GSList *g_slist_find_plugin_name(GSList *list, char *name)
{
	struct tcl_plugin *plug;

	while (list) {
		plug = list->data;
		if (!strcasecmp(plug->name, name))
			return list;
		list = g_slist_next(list);
	}
	return NULL;
}

static GSList *g_slist_find_plugin_interp(GSList *list, Tcl_Interp *interp)
{
	struct tcl_plugin *plug;

	while (list) {
		plug = list->data;
		if (plug->interp == interp)
			return list;
		list = g_slist_next(list);
	}
	return NULL;
}

/*****************************************************************************/
/* Tcl command extensions */

static int tclcmd_print_text(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	struct session *sess;

	if (argc != 3) {
		Tcl_SetResult(interp, "wrong # of args: should be \"print_text session text\"", TCL_STATIC);
		return TCL_ERROR;
	}
	sess = (struct session *) strtoul(argv[1], NULL, 0);
	if (!sess)
		sess = current_tab;
	if (!is_session(sess)) {
		Tcl_SetResult(interp, "session does not exists", TCL_STATIC);
		return TCL_ERROR;
	}
	print_text(sess, argv[2]);
	Tcl_SetResult(interp, "0", TCL_STATIC);
	return TCL_OK;
}

static void do_unregister_plugin(ClientData clientData, Tcl_Interp *interp)
{
	struct tcl_plugin *plug = (struct tcl_plugin *) clientData;
	struct tcl_hook *hook;

	print_text(current_tab, MSG_PREFIX "do_unregister_plugin: 0x%08lx\n", (unsigned long) interp);
	while (plug->hooks) {
		hook = plug->hooks->data;
		plug->hooks = g_slist_remove(plug->hooks, hook);
		unhook_signal(hook->sig);
		g_free(hook->sig);
		g_free(hook);
	}
	tcl_plugins = g_slist_remove(tcl_plugins, plug);
	g_free(plug);
}

static int tclcmd_register_plugin(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	struct tcl_plugin *plug;

	if (argc != 2) {
		Tcl_SetResult(interp, "wrong # of args: should be \"register_plugin plugin_name\"", TCL_STATIC);
		return TCL_ERROR;
	}
	if (g_slist_find_plugin_name(tcl_plugins, argv[1])) {
		Tcl_SetResult(interp, "plugin already exists with that name", TCL_STATIC);
		return TCL_ERROR;
	}
	if (g_slist_find_plugin_interp(tcl_plugins, interp)) {
		Tcl_SetResult(interp, "plugin already registered", TCL_STATIC);
		return TCL_ERROR;
	}

	plug = g_new0(struct tcl_plugin, 1);
	plug->interp = interp;
	plug->hooks = NULL;
	strncpy(plug->name, argv[1], sizeof(plug->name));
	tcl_plugins = g_slist_prepend(tcl_plugins, plug);

	Tcl_CallWhenDeleted(interp, do_unregister_plugin, plug);
	Tcl_SetResult(interp, "0", TCL_STATIC);
	return TCL_OK;
}

static int tclcmd_remote_eval(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	GSList *item;
	struct tcl_plugin *plug;
	int res;

	if (argc != 3) {
		Tcl_SetResult(interp, "wrong # of args: should be \"remote_eval plugin_name command\"", TCL_STATIC);
		return TCL_ERROR;
	}
	if (!(item = g_slist_find_plugin_name(tcl_plugins, argv[1]))) {
		Tcl_SetResult(interp, "plugin does not exist", TCL_STATIC);
		return TCL_ERROR;
	}
	plug = item->data;
	res = Tcl_Eval(plug->interp, argv[2]);
	Tcl_SetResult(interp, Tcl_GetStringResult(plug->interp), TCL_VOLATILE);
	return res;
}

static int tclcmd_hook_signal(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	GSList *item;
	struct tcl_plugin *plug;
	struct tcl_hook *hook;
	int signum = -1, k;
	
	if (argc != 3) {
		Tcl_SetResult(interp, "wrong # of args: should be \"hook_signal signal command\"", TCL_STATIC);
		return TCL_ERROR;
	}
	if (!(item = g_slist_find_plugin_interp(tcl_plugins, interp))) {
		Tcl_SetResult(interp, "plugin not registered", TCL_STATIC);
		return TCL_ERROR;
	}
	plug = item->data;

	for (k = 0; signal_map[k].name; k++)
		if (!strcasecmp(signal_map[k].name, argv[1])) {
			signum = signal_map[k].number;
			break;
		}
	if (signum == -1) {
		Tcl_SetResult(interp, "invalid signal name", TCL_STATIC);
		return TCL_ERROR;
	}

	hook = g_new0(struct tcl_hook, 1);
	hook->plugin = plug;
	strncpy(hook->cmd_name, argv[2], sizeof(hook->cmd_name));
	hook->map_entry = &signal_map[k];
	hook->sig = g_new0(struct xp_signal, 1);
	hook->sig->signal = signum;
	hook->sig->data = hook;
	hook->sig->callback = tcl_stock_signal_handler;
	hook_signal(hook->sig);
	plug->hooks = g_slist_prepend(plug->hooks, hook);

	Tcl_SetResult(interp, "0", TCL_STATIC);
	return TCL_OK;
}

static int tclcmd_tcp_send(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	struct server *serv;
	char rbuf[16], buf[1024];

	if (argc != 3) {
		Tcl_SetResult(interp, "wrong # of args: should be \"tcp_send server string\"", TCL_STATIC);
		return TCL_ERROR;
	}
	serv = (struct server *) strtoul(argv[1], NULL, 0);
	if (!is_server(serv)) {
		Tcl_SetResult(interp, "server does not exist", TCL_STATIC);
		return TCL_ERROR;
	}
	strncpy(buf, argv[2], 1020);
	strcat(buf, "\r\n");

	snprintf(rbuf, 16, "%d", tcp_send(serv, buf));
	Tcl_SetResult(interp, rbuf, TCL_VOLATILE);
	return TCL_OK;
}

static int tclcmd_session(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	struct session *sess;
	char *cmd;

	if (argc < 3) {
		Tcl_SetResult(interp, "wrong # of args: should be \"session session option ?args...?\"", TCL_STATIC);
		return TCL_ERROR;
	}
	sess = (struct session *) strtoul(argv[1], NULL, 0);
	if (!sess)
		sess = current_tab;
	if (!is_session(sess)) {
		Tcl_SetResult(interp, "session does not exist", TCL_STATIC);
		return TCL_ERROR;
	}

	cmd = argv[2];
	if (!strcmp(cmd, "get")) {
		char *var;
		
		if (argc != 4) {
			Tcl_SetResult(interp, "wrong # of args: should be \"session session get variable\"", TCL_STATIC);
			return TCL_ERROR;
		}
		var = argv[3];

#define CK_STR(x) if (!strcmp(var, #x)) { Tcl_SetObjResult(interp, Tcl_NewStringObj(sess->x, strlen(sess->x))); } else
#define CK_INT(x) if (!strcmp(var, #x)) { Tcl_SetObjResult(interp, Tcl_NewIntObj(sess->x)); } else

		CK_STR(channel)
		CK_STR(waitchannel)
		CK_STR(willjoinchannel)
		CK_STR(nick)
		CK_STR(channelkey)
		CK_INT(limit)
		CK_INT(logfd)
		CK_STR(lastnick)
		CK_INT(ops)
		CK_INT(total)
		CK_STR(quitreason)
		CK_INT(is_shell)
		CK_INT(is_tab)
		CK_INT(is_dialog)
		CK_INT(is_server)
		CK_INT(new_data)
		CK_INT(nick_said)
		CK_INT(ignore_date)
		CK_INT(ignore_mode)
		CK_INT(ignore_names)
		CK_INT(end_of_names)
		CK_INT(doing_who)
		CK_INT(userlisthidden)
#ifdef STATUS_MODS
		CK_INT(is_status)
#endif
		if (!strcmp(var, "server")) {
			Tcl_SetObjResult(interp, Tcl_NewLongObj((unsigned long) sess->server));
		} else {
			Tcl_SetResult(interp, "unknown session variable", TCL_STATIC);
			return TCL_ERROR;
		}
#undef CK_STR
#undef CK_INT
		
	} else {
		Tcl_SetResultFmt(interp, "bad option \"%s\": must be get", cmd);
		return TCL_ERROR;
	}
	return TCL_OK;
}

int tclcmd_server(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	struct server *serv;
	char *cmd;

	if (argc < 3) {
		Tcl_SetResult(interp, "wrong # of args: should be \"server server option ?args...?\"", TCL_STATIC);
		return TCL_ERROR;
	}
	serv = (struct server *) strtoul(argv[1], NULL, 0);
	if (!is_server(serv)) {
		Tcl_SetResult(interp, "server does not exist", TCL_STATIC);
		return TCL_ERROR;
	}

	cmd = argv[2];
	if (!strcmp(cmd, "get")) {
		char *var;
		
		if (argc != 4) {
			Tcl_SetResult(interp, "wrong # of args: should be \"server server get variable\"", TCL_STATIC);
			return TCL_ERROR;
		}
		var = argv[3];

#define CK_STR(x) if (!strcmp(var, #x)) { Tcl_SetObjResult(interp, Tcl_NewStringObj(serv->x, strlen(serv->x))); } else
#define CK_INT(x) if (!strcmp(var, #x)) { Tcl_SetObjResult(interp, Tcl_NewIntObj(serv->x)); } else
#define CK_LONG(x) if (!strcmp(var, #x)) { Tcl_SetObjResult(interp, Tcl_NewLongObj(serv->x)); } else
#define CK_TIME(x) if (!strcmp(var, #x)) { Tcl_SetObjResult(interp, Tcl_NewLongObj((unsigned long) serv->x)); } else
#define CK_SESS(x) if (!strcmp(var, #x)) { Tcl_SetObjResult(interp, Tcl_NewLongObj((unsigned long) serv->x)); } else

		CK_INT(port)
		CK_INT(sok)
		CK_INT(childread)
		CK_INT(childwrite)
		CK_INT(childpid)
		CK_INT(iotag)
		CK_INT(bartag)
		CK_STR(hostname)
		CK_STR(servername)
		CK_STR(password)
		CK_STR(nick)
		CK_STR(linebuf)
		CK_LONG(pos)
		CK_INT(nickcount)
		CK_TIME(last_send)
		CK_INT(bytes_sent)
		CK_SESS(front_session)
		CK_INT(motd_skipped)
		CK_INT(connected)
		CK_INT(connecting)
		CK_INT(no_login)
		CK_INT(skip_next_who)
		CK_INT(inside_whois)
		CK_INT(doing_who)
		CK_INT(end_of_motd)
		CK_INT(sent_quit)
		{
			Tcl_SetResult(interp, "unknown server variable", TCL_STATIC);
			return TCL_ERROR;
		}
#undef CK_SESS
#undef CK_TIME
#undef CK_LONG
#undef CK_STR
#undef CK_INT
		
	} else {
		Tcl_SetResultFmt(interp, "bad option \"%s\": must be get", cmd);
		return TCL_ERROR;
	}
	return TCL_OK;
}

/*****************************************************************************/
/* Tcl interpreters */

static Tcl_Interp *tcl_new_interp(void)
{
	Tcl_Interp *interp;

	if (!(interp = Tcl_CreateInterp()))
		return NULL;
	if (Tcl_Init(interp) == TCL_ERROR)
		return NULL;
	Tcl_Preserve(interp);

	Tcl_SetStdChannel(Tcl_CreateChannel(&mystream_type, 
		"stdin", 0, TCL_READABLE), TCL_STDIN);
	Tcl_SetStdChannel(Tcl_CreateChannel(&mystream_type, 
		"stdout", 0, TCL_WRITABLE), TCL_STDOUT);
	Tcl_SetStdChannel(Tcl_CreateChannel(&mystream_type, 
		"stderr", 0, TCL_WRITABLE), TCL_STDERR);

	Tcl_CreateCommand(interp, "print_text", tclcmd_print_text, NULL, NULL);
	Tcl_CreateCommand(interp, "hook_signal", tclcmd_hook_signal, NULL, NULL);
	Tcl_CreateCommand(interp, "register_plugin", tclcmd_register_plugin, NULL, NULL);
	Tcl_CreateCommand(interp, "remote_eval", tclcmd_remote_eval, NULL, NULL);
	Tcl_CreateCommand(interp, "tcp_send", tclcmd_tcp_send, NULL, NULL);
	Tcl_CreateCommand(interp, "session", tclcmd_session, NULL, NULL);
	Tcl_CreateCommand(interp, "server", tclcmd_server, NULL, NULL);

	return interp;
}

static void tcl_kill_interp(Tcl_Interp *interp)
{
	if (!Tcl_InterpDeleted(interp))
		Tcl_DeleteInterp(interp);
	Tcl_Release(interp);
}

/*****************************************************************************/

/* globals */
Tcl_Interp *main_interp;

int xcmd_tcl(struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	if (!word_eol[2] || !word_eol[2][0])
		return 1;
	if (Tcl_VarEval(main_interp, word_eol[2], "\n", NULL) != TCL_OK)
		print_text(sess, MSG_PREFIX "error: %s", Tcl_GetStringResult(main_interp));
	return 0;
}

int xcmd_loadtcl(struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	Tcl_Interp *interp;

	if (!word[2] || !word[2][0])
		return 1;
	if (!(interp = tcl_new_interp())) {
		print_text(sess, MSG_PREFIX "cannot create Tcl interpreter\n");
		return 0;
	}
	if (Tcl_EvalFile(interp, word[2]) != TCL_OK)
		print_text(sess, MSG_PREFIX "cannot load script: %s", Tcl_GetStringResult(interp));
	if (!g_slist_find_plugin_interp(tcl_plugins, interp))
		tcl_kill_interp(interp);
	return 0;
}

int xcmd_rmtcl(struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	GSList *item;
	struct tcl_plugin *plug;

	if (!word[2] || !word[2][0])
		return 1;
	if (!(item = g_slist_find_plugin_name(tcl_plugins, word[2]))) {
		print_text(sess, MSG_PREFIX "plugin %s not found", word[2]);
		return 0;
	}
	plug = item->data;
	tcl_kill_interp(plug->interp);
	return 0;
}

int xcmd_debugtcl(struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	GSList *item, *subitem;
	struct tcl_plugin *plug;
	struct tcl_hook *hook;

	for (item = tcl_plugins; item; item = g_slist_next(item)) {
		plug = item->data;
		print_text(sess, "Plugin: 0x%08lx %s",
			(unsigned long) plug->interp,
			plug->name);
		for (subitem = plug->hooks; subitem; subitem = g_slist_next(subitem)) {
			hook = subitem->data;
			print_text(sess, "  Hook: %d %s",
				hook->sig->signal,
				hook->cmd_name);
		}
	}
	return 0;
}

struct commands tcl_cmds[] = {
	{ "TCL",	xcmd_tcl,	0, 0, "/TCL <raw tcl command>" },
	{ "LOADTCL",	xcmd_loadtcl,	0, 0, "/LOADTCL <filename.tcl>" },
	{ "RMTCL",	xcmd_rmtcl,	0, 0, "/RMTCL <plugin>" },
	{ "DEBUGTCL",	xcmd_debugtcl,	0, 0, "/DEBUGTCL" },
	{ NULL,		NULL,		0, 0, NULL }
};

struct module_cmd_set tcl_cmd_set;

int module_init(int ver, struct module *mod, struct session *sess)
{
	/* This check *MUST* be done first */
	if (ver != MODULE_IFACE_VER)
		return 0;
	
	if (module_find(mod_name)) {
		print_text(sess, MSG_PREFIX "module already loaded\n");
		return 0;
	}
	mod->name = mod_name;
	mod->desc = mod_desc;

	if (!(main_interp = tcl_new_interp())) {
		print_text(sess, MSG_PREFIX "cannot create Tcl interpreter\n");
		return 1;
	}

	tcl_cmd_set.mod = mod;
	tcl_cmd_set.cmds = tcl_cmds;
	module_add_cmds(&tcl_cmd_set);
	print_text(sess, MSG_PREFIX "module loaded\n");
	return 0;
}

void module_cleanup(struct module *mod, struct session *sess)
{
	struct tcl_plugin *plug;

	while (tcl_plugins) {
		plug = tcl_plugins->data;
		tcl_kill_interp(plug->interp);
	}

	tcl_kill_interp(main_interp);
	print_text(sess, MSG_PREFIX "module unloaded\n");
}
